/*******************************************************************************
 * Copyright (c) 2013 Piotr Piastucki
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package org.vcs.bazaar.eclipse.core.commands;

import java.io.File;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.vcs.bazaar.client.BazaarStatusKind;
import org.vcs.bazaar.client.BazaarTreeStatus;
import org.vcs.bazaar.client.IBazaarClient;
import org.vcs.bazaar.client.IBazaarItemInfo;
import org.vcs.bazaar.client.IBazaarStatus;
import org.vcs.bazaar.client.commandline.commands.options.Option;
import org.vcs.bazaar.client.commandline.syntax.ILsOptions;
import org.vcs.bazaar.client.core.BazaarClientException;
import org.vcs.bazaar.eclipse.BzrWorkspaceRoot;
import org.vcs.bazaar.eclipse.core.model.IBzrResourceStatus;
import org.vcs.bazaar.eclipse.core.model.local.IBzrLocalResource;
import org.vcs.bazaar.eclipse.core.utils.ResourceUtil;
import org.vcs.bazaar.eclipse.internal.core.BazaarException;
import org.vcs.bazaar.eclipse.internal.core.model.BzrResourceStatus;

/**
 * This class is invoked form StatusCacheManager to build status map that is later used to decorate resources. 
 * @author Guillermo Gonzalez
 * @author Piotr Piastucki
 */
public class StatusUpdaterCommand extends BazaarCommand {

	private final IBzrLocalResource bzrLocalResource;
	private Map<IResource, IBzrResourceStatus> statusMap;
	private Option[] options = new Option[0];
	private final boolean shallow;

	public StatusUpdaterCommand(IBzrLocalResource resource, boolean shallow) {
		this.bzrLocalResource = resource;
		this.shallow = shallow;
	}

	@Override
    public void run(IProgressMonitor monitor) throws BazaarException {
		try {
			statusMap = new HashMap<IResource, IBzrResourceStatus>();
			IBazaarClient client = getClient(bzrLocalResource.getIResource().getProject(), true);
			executeAndFilterLogOutput(client);
		} catch (BazaarClientException e) {
			throw BazaarException.wrapException(e, bzrLocalResource.getIResource());
		} catch (BazaarException e) {
			throw e;
		} catch (CoreException e) {
			throw BazaarException.wrapException(e);
		}
	}

	public Map<IResource, IBzrResourceStatus> getStatus() {
		return statusMap;
	}

	private void executeAndFilterLogOutput(IBazaarClient client) throws BazaarClientException, CoreException {
		Set<IResource> versionedSet = new HashSet<IResource>();
		Set<IResource> ignoredSet = new HashSet<IResource>();
		Set<IResource> unknownSet = new HashSet<IResource>();
		IResource resourceToRefresh = bzrLocalResource.getIResource();
		File file = resourceToRefresh.getLocation().toFile();
		/* get all versioned and ignored resources with ls command excluding derived resources */
		IBazaarItemInfo[] paths;
		if (shallow) {
			paths = client.ls(file, null, ILsOptions.IGNORED, ILsOptions.VERSIONED, ILsOptions.UNKNOWN, ILsOptions.NON_RECURSIVE);
		} else {
			paths = client.ls(file, null, ILsOptions.IGNORED, ILsOptions.VERSIONED, ILsOptions.UNKNOWN);
		}
		if (paths != null) {
			for (IBazaarItemInfo itemInfo : paths) {
				Path path = new Path(itemInfo.getPath());
				int matchingSegments = resourceToRefresh.getProject().getLocation().matchingFirstSegments(path);
				IResource resource = resourceToRefresh.getProject().findMember(path.removeFirstSegments(matchingSegments));
				if (resource != null) {
					if (ResourceUtil.shouldHandle(resource, false)) {
						if (itemInfo.getStatusKind() == BazaarStatusKind.IGNORED) {
							ignoredSet.add(resource);
						} else if (itemInfo.getStatusKind() == BazaarStatusKind.UNKNOWN) {
							unknownSet.add(resource);
						} else {
							versionedSet.add(resource);
						}
					}
				}
			}
			if (resourceToRefresh.getType() == IResource.PROJECT && !versionedSet.contains(resourceToRefresh)) {
				versionedSet.add(resourceToRefresh);
			}
		}
		/* execute st command and turn its result into a status map */
		BazaarTreeStatus treeStatus = client.status(new File[]{file}, options);
		IBazaarStatus[] statuses = treeStatus.getStatusAsArray();
		/* add everything returned by bzr status command excluding derived resources */
		for (IBazaarStatus aStatus : statuses) {
			IResource resource = ResourceUtil.getIResource(aStatus);
			if (resource != null) {
				if (ResourceUtil.shouldHandle(resource, false)) {
					IBzrResourceStatus resourceStatus = new BzrResourceStatus(BzrWorkspaceRoot.getBzrResourceFor(resource), aStatus);
					statusMap.put(resource, resourceStatus);
					if (resourceStatus.isUnknown() && resource instanceof IContainer) {
						unknownSet.add(resource);
					}
				}
			}
		}
		/* add all files and sub-folders inside IGNORED folders with IGNORED status, it needs to be done manually as 
		 * ignored resources are not reported recursively by bzr ls command */
		for (IResource ignored : ignoredSet) {
			ignored.accept(new IResourceVisitor() {
				public boolean visit(IResource resource) throws CoreException {
					if (!statusMap.containsKey(resource)) {
						statusMap.put(resource, BzrResourceStatus.ignored(resource));
					}
					return true;
				}
			}, IResource.DEPTH_INFINITE, IFolder.EXCLUDE_DERIVED);
		}
		/* add all files and sub-folders inside UNKNOWN folders with UNKNOWN status, it needs to be done manually as 
		 * untracked resources are not reported recursively by bzr status command */
		for (IResource unknown : unknownSet) {
			unknown.accept(new IResourceVisitor() {
				public boolean visit(IResource resource) throws CoreException {
					if (!statusMap.containsKey(resource)) {
						statusMap.put(resource, BzrResourceStatus.unknown(resource));
					}
					return true;
				}
			}, IResource.DEPTH_INFINITE, IFolder.EXCLUDE_DERIVED);
		}
		for (IResource versioned : versionedSet) {
			if (!statusMap.containsKey(versioned)) {
				statusMap.put(versioned, BzrResourceStatus.unchanged(versioned));
			}
		}
	}
	
	/* 
	 * Noticed bzr bugs:
	 * 1) Ignored resources (changes not yet committed) are listed as unknown by bzr ls 
	 *    i.e. bzr ls --unknown shows the resources while bzr ls --ignored does not.
	 *    Unfortunately, bzr st does not list such resources at all. 
	 */
	
}
