/**
 * LICENSE + COPYRIGHT
 */
package org.vcs.bazaar.eclipse.core.resources;

import java.io.IOException;
import java.io.InputStream;

import org.eclipse.core.resources.IEncodedStorage;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.PlatformObject;
import org.eclipse.core.runtime.content.IContentDescription;
import org.eclipse.core.runtime.content.IContentTypeManager;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.core.BazaarClientException;
import org.vcs.bazaar.client.core.BranchLocation;
import org.vcs.bazaar.eclipse.BzrWorkspaceRoot;
import org.vcs.bazaar.eclipse.EclipseBazaarCore;
import org.vcs.bazaar.eclipse.core.commands.CatCommand;
import org.vcs.bazaar.eclipse.core.model.IBzrBranch;
import org.vcs.bazaar.eclipse.internal.core.BazaarException;

/**
 * @author zingo
 * @modifiedby Guillermo Gonzalez
 *
 * This is a IStorage subclass that handle file revision
 *
 */
public class IStorageBazaarRevision extends PlatformObject implements IEncodedStorage {
	private final IBzrBranch branch;
	private final IResource resource;
	private final BazaarRevision revision;
	private InputStream contents;
	private BranchLocation location;

	/**
	 * This constructor is used to create storage with remote content
	 * @param resource
	 * @param branch
	 * @param relativePath
	 * @param revision
	 */
	public IStorageBazaarRevision(IResource resource, IBzrBranch branch, String relativePath, BazaarRevision revision) {
		this.resource = resource;
		this.revision = revision;
		this.branch = branch;
		String path;
		if (branch.getPath() != null) {
			path = branch.getPath().append(relativePath).toOSString();
		} else {
			path = branch.getBranchLocation() + "/" + relativePath;
		}
		try {
			location = new BranchLocation(path);
		} catch (BazaarClientException e) {
			EclipseBazaarCore.log(BazaarException.wrapException(e));
		}
	}

	public IStorageBazaarRevision(IResource resource, BazaarRevision revision) {
		this.branch = BzrWorkspaceRoot.getBranchFor(resource.getProject());
		this.resource = resource;
		this.revision = revision;
	}

	public IStorageBazaarRevision(IResource resource) {
		this(resource, getLastRevision(resource));
	}

	private static BazaarRevision getLastRevision(IResource resource) {
		BazaarRevision rev = null;
		try {
			rev = EclipseBazaarCore.getDefault().getStatusCacheManager().getResourceRevision(resource);
		} catch (BazaarException e) {
			EclipseBazaarCore.log(e);
		}
		return rev;
	}

	@Override
    @SuppressWarnings("unchecked")
	public Object getAdapter(Class adapter) {
		return null;
	}

	public InputStream getContents() throws CoreException {
		if (resource.getType() != IResource.FILE) {
			return null;
		}
		// store the inputStream in a private member so we don't need to call 'bzr cat' again
		if (contents == null) {
			String charsetName;
			try {
				charsetName = ((IFile)resource).getCharset();
			} catch (CoreException e) {
				charsetName = null;
			}
			CatCommand cmd = location == null ? new CatCommand(branch, resource.getLocation().toFile(), charsetName, revision) : new CatCommand(branch, location, charsetName, revision);
			cmd.run(new NullProgressMonitor());
			contents = cmd.getContent();
		} else {
			// if the inputStream has been already initialized, reset it to the
			// initial position so the client can get all the contents
			try {
				contents.reset();
			} catch (IOException e) {
				throw BazaarException.wrapException(e);
			}
		}
		return contents;
	}

	public IPath getFullPath() {
		return resource.getFullPath();
	}

	public String getBranchLocation() {
		return branch.getBranchLocation().toString();
	}
	
	public String getName() {
		final StringBuilder sb = new StringBuilder();
		sb.append(resource.getName());
		sb.append(" ");
		sb.append(revision==null?"last":revision.toString());
		return sb.toString();
	}

	public boolean isReadOnly() {
		// You can't write to old revisions e.g. ReadOnly
		return true;
	}

	public IResource getResource() {
		return resource;
	}

	public BazaarRevision getRevision() {
		return revision;
	}

	public String getResourceCharset() throws CoreException {
		// tries to obtain a description for this file contents
		String charset = null;
		try {
			IContentTypeManager contentTypeManager = Platform.getContentTypeManager();
			charset = contentTypeManager.getDescriptionFor(getContents(), getName(), IContentDescription.ALL).getCharset();
		} catch (Exception e) {
		}
		if(charset == null && resource.getType() == IResource.FILE) {
			charset = ((IFile)resource).getCharset();
		} else if(charset == null){
			// fallback to parent charset
			return resource.getParent().getDefaultCharset();
		}
		return charset;
	}

	public String getCharset() throws CoreException {
		return getResourceCharset();
	}

}
