/*******************************************************************************
 * Copyright (c) 2013 Piotr Piastucki
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package org.vcs.bazaar.eclipse.core.synchronize;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.IPath;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.BazaarStatusKind;
import org.vcs.bazaar.client.BazaarTreeStatus;
import org.vcs.bazaar.client.IBazaarLogMessage;
import org.vcs.bazaar.client.IBazaarStatus;
import org.vcs.bazaar.eclipse.core.synchronize.BazaarSyncCacheEntry.ChangeType;
import org.vcs.bazaar.eclipse.core.synchronize.BazaarSyncCacheEntry.Direction;

/**
 * @author Piotr Piastucki
 */
public class BazaarSyncCache {

	private Map<IResource, BazaarSyncCacheEntry> cache;

	private static class ChangeAndRevision {
		private final ChangeType changeType;
		private final BazaarRevision revision;
		private BazaarRevision baseRevision;
		public ChangeAndRevision(ChangeType changeType, BazaarRevision revision) {
			this.changeType = changeType;
			this.revision = revision;
			this.baseRevision = revision;
		}
		public ChangeType getChangeType() {
			return changeType;
		}
		public BazaarRevision getRevision() {
			return revision;
		}
		public BazaarRevision getBaseRevision() {
			return baseRevision;
		}
		public void setBaseRevision(BazaarRevision baseRevision) {
			this.baseRevision = baseRevision;
		}
	}
	
	public BazaarSyncCache(IPath branchPath, List<IBazaarLogMessage> mine, List<IBazaarLogMessage> other, BazaarTreeStatus status, BazaarRevision currentRevision) {
		Map<String, ChangeAndRevision> incoming = getChanges(other);
		Map<String, ChangeAndRevision> outgoing = getChanges(mine);
		addStatuses(outgoing, status.getStatus(), currentRevision);
		cache = new HashMap<IResource, BazaarSyncCacheEntry>();
		for (BazaarSyncCacheEntry e : mergeChanges(incoming, outgoing, currentRevision)) {
			IResource resource = ResourcesPlugin.getWorkspace().getRoot().getFileForLocation(branchPath.append(e.getPath()));
			if (resource != null) {
				cache.put(resource, e);
			}
		}
	}

	public BazaarSyncCacheEntry getEntry(IResource resource) {
		return cache.get(resource);
	}
	/**
	 * Merge incoming and outgoing changes.
	 */
	private Collection<BazaarSyncCacheEntry> mergeChanges(Map<String, ChangeAndRevision> incoming, Map<String, ChangeAndRevision> outgoing, BazaarRevision currentRevision) {
		Set<BazaarSyncCacheEntry> entries = new HashSet<BazaarSyncCacheEntry>();
		Set<String> keys = new HashSet<String>(incoming.keySet());
		keys.addAll(outgoing.keySet());
		int maxLocalRevision = Integer.parseInt(currentRevision.getValue());
		for (String key : keys) {
			ChangeAndRevision localChange = outgoing.get(key);
			ChangeAndRevision remoteChange = incoming.get(key);
			BazaarSyncCacheEntry entry = null;
			if (localChange == null) {
				int remoteBase = Integer.parseInt(remoteChange.getBaseRevision().getValue());
				BazaarRevision rev = BazaarRevision.getRevision(Math.max(1, Math.min(maxLocalRevision, remoteBase - 1)));
				entry = new BazaarSyncCacheEntry(key, Direction.INCOMING, remoteChange.getChangeType(), rev, remoteChange.getRevision());
			} else if (remoteChange == null) {
				int localBase = Integer.parseInt(localChange.getBaseRevision().getValue());
				BazaarRevision rev = BazaarRevision.getRevision(Math.max(1, localBase - 1));
				entry = new BazaarSyncCacheEntry(key, Direction.OUTGOING, localChange.getChangeType(), localChange.getRevision(), rev);
			} else if (localChange.getChangeType() == ChangeType.DELETE && remoteChange.getChangeType() == ChangeType.DELETE) {
				continue;
			} else {
				int localBase = Integer.parseInt(localChange.getBaseRevision().getValue());
				int remoteBase = Integer.parseInt(remoteChange.getBaseRevision().getValue());
				BazaarRevision rev = BazaarRevision.getRevision(Math.max(1, Math.min(localBase, remoteBase) - 1));
				entry = new BazaarSyncCacheEntry(key, Direction.CONFLICTING, 
						localChange.getChangeType() == remoteChange.getChangeType() ? localChange.getChangeType() : ChangeType.MODIFY,
						rev, remoteChange.getRevision());
			}
			entries.add(entry);
		}
		return entries;
	}

	/**
	 * Get all affected files in all the given revisions and build a map of changes
	 * @param logs
	 * @return
	 */
	private Map<String, ChangeAndRevision> getChanges(List<IBazaarLogMessage> logs) {
		Map<String, ChangeAndRevision> changes = new HashMap<String, ChangeAndRevision>();
		if (logs != null) {
			for (IBazaarLogMessage log : logs) {
				Collection<IBazaarStatus> statuses = log.getAffectedFiles();
				addStatuses(changes, statuses, log.getRevision());
			}
		}
		return changes;
	}

	/**
	 * Add changes for a single resource to a map and set base revision to the earliest revision affecting the given resource
	 * @param changes
	 * @param statuses
	 * @param revision
	 */
	private void addStatuses(Map<String, ChangeAndRevision> changes, Collection<IBazaarStatus> statuses, BazaarRevision revision) {
		for (IBazaarStatus status : statuses) {
			String path = status.getPath();
			ChangeType changeType = null;
			if (status.contains(BazaarStatusKind.CREATED)) {
				changeType = ChangeType.ADD;
			} else if (status.contains(BazaarStatusKind.DELETED)) {
				changeType = ChangeType.DELETE;
			} else {
				changeType = ChangeType.MODIFY;
			}
			if (changeType == ChangeType.ADD || !changes.containsKey(path)) {
				changes.put(path, new ChangeAndRevision(changeType, revision));
			} else {
				changes.get(path).setBaseRevision(revision);
			}
		}
	}

	/**
	 * Get all affected resources
	 * @return
	 */
	public IResource[] getChangedResources() {
		return cache.keySet().toArray(new IResource[cache.keySet().size()]);
	}

}