/*******************************************************************************
 * Copyright (c) 2013 Piotr Piastucki
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package org.vcs.bazaar.eclipse.ui.actions;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.IJobStatus;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.window.Window;
import org.eclipse.ui.actions.WorkspaceModifyOperation;
import org.eclipse.ui.progress.UIJob;
import org.vcs.bazaar.client.commandline.commands.options.Option;
import org.vcs.bazaar.client.commandline.syntax.IShelveOptions;
import org.vcs.bazaar.eclipse.BzrWorkspaceRoot;
import org.vcs.bazaar.eclipse.CoreTexts;
import org.vcs.bazaar.eclipse.EclipseBazaarCore;
import org.vcs.bazaar.eclipse.core.commands.ShelveCommand;
import org.vcs.bazaar.eclipse.core.model.IBzrResourceStatus;
import org.vcs.bazaar.eclipse.core.model.local.IBzrLocalResource;
import org.vcs.bazaar.eclipse.core.utils.BzrJobStatus;
import org.vcs.bazaar.eclipse.internal.core.BazaarException;
import org.vcs.bazaar.eclipse.ui.EclipseBazaarUI;
import org.vcs.bazaar.eclipse.ui.UITexts;
import org.vcs.bazaar.eclipse.ui.actions.CommitAction.CommitableResources;
import org.vcs.bazaar.eclipse.ui.dialogs.ShelveDialog;

/**
 * @author Piotr Piastucki
 */
public class ShelveAction extends WorkbenchAction {

	@Override
	protected void execute(IAction action) throws Exception {
		final IProject[] projects = getEnclosingProjectsForSelectedResources();
		final ShelveDialog shelveDialog = new ShelveDialog(getShell());
		UIJob job = new UIJob(getShell().getDisplay(), UITexts.ShelveAction_scanningTaskName) {
			@Override
			public IStatus runInUIThread(IProgressMonitor monitor) {
				monitor.beginTask(UITexts.ShelveAction_scanningTaskName, 1000);
				try {
					CommitableResources resourceStauses = CommitAction.createCommitableResources(monitor,
							getEnclosingProjectsForSelectedResources());
					CommitableResources selectedResource = CommitAction.createCommitableResources(monitor,
							getSelectedResources());
					monitor.worked(1000);
					if (resourceStauses.getAllStatuses().isEmpty()) {
						return BzrJobStatus.INFO(this, UITexts.ShelveAction_nothingToShelve);
					}
					List<CommitableResources> results = new ArrayList<CommitableResources>(2);
					results.add(resourceStauses);
					results.add(selectedResource);
					return new BzrJobStatus<List<CommitableResources>>(IJobStatus.OK, this, results,
							"Resources calculated");
				} catch (Exception e) {
					return UIJob.errorStatus(e);
				} finally {
					monitor.done();
				}
			}
		};
		job.setUser(true);
		job.setPriority(UIJob.LONG);
		job.addJobChangeListener(new JobChangeAdapter() {
			@SuppressWarnings("unchecked")
			@Override
			public void done(IJobChangeEvent event) {
				IStatus status = event.getResult();
				if (status instanceof BzrJobStatus) {
					if (status.getSeverity() == IStatus.INFO) {
						MessageDialog.openInformation(getParentShell(), UITexts.Dialog_informationTitle, status.getMessage());
					} else if (status.getSeverity() == IStatus.OK) {
						List<CommitableResources> result = ((BzrJobStatus<List<CommitableResources>>) status)
								.getResult();
						int exit = shelveDialog.open(result.get(0).getAllStatuses(), result.get(1).getAllStatuses(),
								true);
						if (exit == Window.OK) {
							try {
								handleSelected(result.get(0).getAllStatuses(), projects, shelveDialog);
							} catch (Exception e) {
								handle(e);
							}
						}
					}
				}
				event.getJob().removeJobChangeListener(this);
			}
		});
		job.schedule();
	}

	private void handleSelected(final List<IBzrResourceStatus> commitableResources, final IProject[] projects,
			ShelveDialog shelveDialog) throws BazaarException, InvocationTargetException, InterruptedException {

		final List<IBzrResourceStatus> filesToShelve = shelveDialog.getFilesToShelve();
		final String message = shelveDialog.getMessage();
		final boolean wholeBranch = shelveDialog.isWholeBranch();
		final List<Option> options = new ArrayList<Option>();
		options.add(IShelveOptions.ALL);
		if (message != null && !message.trim().isEmpty()) {
			options.add(IShelveOptions.MESSAGE.with(message));
		}
		run(new WorkspaceModifyOperation() {
			@Override
            public void execute(IProgressMonitor monitor) throws InvocationTargetException {
				try {
					monitor.beginTask(CoreTexts.ShelveCommand_taskName, 1000);
					monitor.setTaskName(CoreTexts.ShelveCommand_taskName);
					final List<IBzrLocalResource> resourcesToShelve = new ArrayList<IBzrLocalResource>();
					if (wholeBranch) {
						resourcesToShelve.addAll(Arrays.asList(BzrWorkspaceRoot.getBzrResourcesFor(projects)));
					} else {
						for (IBzrResourceStatus bzrResourceStatus : filesToShelve) {
							resourcesToShelve.add(bzrResourceStatus.getResource());
						}
					}
					final ShelveCommand cmd = new ShelveCommand(resourcesToShelve, wholeBranch, options.toArray(new Option[0]));
					cmd.run(EclipseBazaarCore.subProgressMonitorFor(monitor));
				} catch (Exception e) {
					EclipseBazaarUI.log(e.getMessage(), e);
				} finally {
					monitor.done();
				}
			}
		}, false /* no cancelable */, PROGRESS_DIALOG);
	}

	@Override
	protected boolean isEnabledForUnmanagedResources() {
		return false;
	}

	@Override
	protected boolean isEnabledForUnmodifiedResources() {
		if (getSelectedProjects().length > 0) {
			return true;
		}
		return false;
	}

	@Override
	protected boolean isEnabledForMultipleEclosingProjects() {
		return isSameBranchInEnclosingProjects();
	}

}
