/*******************************************************************************
 * Copyright (c) 2013 Piotr Piastucki
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package org.vcs.bazaar.eclipse.ui.compare;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.compare.CompareConfiguration;
import org.eclipse.compare.CompareEditorInput;
import org.eclipse.compare.IContentChangeListener;
import org.eclipse.compare.IContentChangeNotifier;
import org.eclipse.compare.IEditableContent;
import org.eclipse.compare.ResourceNode;
import org.eclipse.compare.structuremergeviewer.ICompareInput;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.ui.IWorkbenchPage;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.eclipse.core.model.IBzrResourceStatus;
import org.vcs.bazaar.eclipse.core.model.local.IBzrLocalResource;
import org.vcs.bazaar.eclipse.core.resources.IStorageBazaarRevision;

/**
 * @author Piotr Piastucki
 */
public class BazaarStatusBasedCompareEditorInput extends CompareEditorInput implements IContentChangeListener {

	private IResource[] resources;
	private List<IBzrResourceStatus> statuses;
	private BazaarRevision leftRevision;
	private BazaarRevision rightRevision;
	private boolean readOnly;
	private List<BazaarLocalResourceNode> changedNodes;
	
	public BazaarStatusBasedCompareEditorInput(IResource[] resources, List<IBzrResourceStatus> statuses,
			BazaarRevision leftRevision, BazaarRevision rightRevision, IWorkbenchPage page) {
		super(new CompareConfiguration());
		this.resources = resources;
		this.statuses = statuses;
		this.leftRevision = leftRevision;
		this.rightRevision = rightRevision;
		readOnly = leftRevision != null;
		changedNodes = new ArrayList<BazaarLocalResourceNode>();
	}

	@Override
	protected ICompareInput prepareInput(IProgressMonitor monitor)
			throws InvocationTargetException, InterruptedException {
		ICompareInput input = createCompareInput();
		getCompareConfiguration().setRightEditable(false);
		getCompareConfiguration().setLeftEditable(isLeftEditable(input));
		initLabels(input);
		return input;
	}

	@Override
	public void saveChanges(IProgressMonitor monitor) throws CoreException {
		flushViewers(monitor);
		List<BazaarLocalResourceNode> nodes = new ArrayList<BazaarLocalResourceNode>(changedNodes);
		changedNodes.clear();
		for (ResourceNode node : nodes) {
			if (node instanceof BazaarLocalResourceNode) {
				((BazaarLocalResourceNode)node).commit(monitor);
			}
		}
	}
	
	private ICompareInput createCompareInput() {
		BazaarResourceDiffNodeTree tree = new BazaarResourceDiffNodeTree(readOnly, resources, this);
		for (IBzrResourceStatus local : statuses) {
			ResourceNode ln = null;
			ResourceNode rn = null;
			IBzrLocalResource localResource = local.getResource();
			if (!local.isDeleted()) {
				ln = leftRevision == null ? 
					new BazaarLocalResourceNode(localResource) :
					new BazaarRevisionNode(new IStorageBazaarRevision(localResource.getIResource(), leftRevision));
				ln.addContentChangeListener(this);
			}
			if (!(local.isAdded() || local.isUnversioned() || local.isUnknown())) {
				rn = rightRevision == null ? 
					new BazaarRevisionNode(new IStorageBazaarRevision(localResource.getIResource())) :
					new BazaarRevisionNode(new IStorageBazaarRevision(localResource.getIResource(), rightRevision));
			}
			if (ln != null || rn != null) {
				tree.add(ln, rn);
			}
		}
		return tree.getNodes();
	}

	private void initLabels(ICompareInput input) {
		CompareEditorUtil.setLabels(this, input.getLeft(), input.getRight(), leftRevision, rightRevision);
	}

	private boolean isLeftEditable(ICompareInput input) {
		Object left = input.getLeft();
		if (left instanceof IEditableContent) {
			return ((IEditableContent) left).isEditable();
		}
		return !readOnly;
	}

	public void contentChanged(IContentChangeNotifier source) {
		if (source instanceof BazaarLocalResourceNode) {
			changedNodes.add((BazaarLocalResourceNode) source);
		}
	}

}
