/**
 * LICENSE + COPYRIGHT
 */
package org.vcs.bazaar.eclipse.ui.dialogs;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.eclipse.jface.dialogs.Dialog;
import org.eclipse.jface.dialogs.IInputValidator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.FocusEvent;
import org.eclipse.swt.events.FocusListener;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.events.SelectionListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Shell;
import org.vcs.bazaar.client.core.BazaarClientException;
import org.vcs.bazaar.client.core.BranchLocation;
import org.vcs.bazaar.eclipse.core.model.IBzrBranch;
import org.vcs.bazaar.eclipse.core.repository.BranchFactory;
import org.vcs.bazaar.eclipse.internal.core.model.RemoteBranch;
import org.vcs.bazaar.eclipse.ui.UITexts;
import org.vcs.bazaar.eclipse.ui.widgets.RevisionNumber;

/**
 * @author Guillermo Gonzalez
 *
 */
public class LocationInputDialog extends InputDialog {

	private Combo uriCombo;
	private Button remember;
	private Button force;
	private Collection<IBzrBranch> locations;
	private LocationDialogInfo info;
	private Button useExistingDir;
	private boolean optionRemember;
	private boolean optionForce;
	private boolean optionUseExistingDir;
	private RevisionNumber revision;
	
	protected static class BaseURIValidator implements IInputValidator {
		public String isValid(String newText) {
			try {
				new BranchLocation(newText);
			} catch (BazaarClientException e) {
				return "Please enter a valid branch location";
			}
			return null;
		}
	};

	public static LocationInputDialog remoteBindLocationDialog(Shell parentShell, Collection<IBzrBranch> locations, IBzrBranch initialBranch, LocationDialogInfo info) {
		return remoteLocationDialog(parentShell, locations, initialBranch, info, "Please enter the URI to bind", false);
	}

	public static LocationInputDialog remotePushLocationDialog(Shell parentShell, Collection<IBzrBranch> locations, IBzrBranch initialBranch, LocationDialogInfo info) {
		return remoteLocationDialog(parentShell, locations, initialBranch, info, "Please enter the URI to push to", true, false, true);
	}

	public static LocationInputDialog remotePullLocationDialog(Shell parentShell, Collection<IBzrBranch> locations,IBzrBranch initialBranch,  LocationDialogInfo info) {
		return remoteLocationDialog(parentShell, locations, initialBranch, info, "Please enter the URI to pull from", true);
	}

	public static LocationInputDialog missingLocationDialog(Shell parentShell, Collection<IBzrBranch> locations, LocationDialogInfo info) {
		return remoteLocationDialog(parentShell, locations, null, info, "Please enter the URI to look for missing revisions", false);
	}

	/**
	 *
	 * @param parentShell
	 * @param locations
	 * @param info
	 * @param dialogMessage
	 * @param rememberOption if true shows a checkbox fo remember option in pull/push actions
	 * @return
	 * LocationInputDialog
	 */
	public static LocationInputDialog remoteLocationDialog(Shell parentShell, Collection<IBzrBranch> locations, IBzrBranch initialBranch, LocationDialogInfo info, String dialogMessage, boolean remember) {
		return remoteLocationDialog(parentShell, locations, initialBranch, info, dialogMessage, remember, false, false);
	}

	public static LocationInputDialog remoteLocationDialog(Shell parentShell, Collection<IBzrBranch> locations, IBzrBranch initialBranch, LocationDialogInfo info, String dialogMessage, boolean remember, boolean force, boolean useExistingDir) {
		return new LocationInputDialog(parentShell, "Remote Location", dialogMessage, initialBranch, new BaseURIValidator(), locations, info, remember, force, useExistingDir);
	}

	public static LocationInputDialog switchLocationDialog(Shell parentShell, Collection<IBzrBranch> locations, LocationDialogInfo dialogInfo) {
		return new LocationInputDialog(parentShell, "Switch Location", "Please enter the URI to switch to", null, new BaseURIValidator(), locations, dialogInfo, false, true, false);
	}

	public LocationInputDialog(Shell parentShell, String dialogTitle, String dialogMessage, IBzrBranch initialBranch, IInputValidator validator, Collection<IBzrBranch> locations, LocationDialogInfo info, boolean remember, boolean force, boolean useExistingDir) {
		super(parentShell, dialogTitle, dialogMessage, getBranchString(initialBranch), validator);
		this.locations = locations;
		this.info = info;
		this.optionRemember = remember;
		this.optionForce = force;
		this.optionUseExistingDir = useExistingDir;
	}

	protected static String getBranchString(IBzrBranch branch) { 
		String value = "";
		if (branch != null && branch.getBranchLocation() != null) {
			value = branch.getBranchLocation().toString();
		}
		return value;
	}
	
	protected LocationInputDialog(Shell shell, String dialogTitle, String dialogMessage, IBzrBranch initialBranch, BaseURIValidator validator) {
		super(shell, dialogTitle, dialogMessage, getBranchString(initialBranch), validator);
		this.locations = Collections.emptyList();
		this.info = new LocationDialogInfo();
	}

	public void loadBranchesInCombo() {
		for (IBzrBranch branch : locations) {
			if (branch.getBranchLocation() != null) {
				uriCombo.add(branch.getBranchLocation().toString());
			}
		}
	}

	@Override
    protected void createTextInput(Composite composite) {
		uriCombo = new Combo(composite, SWT.SINGLE | SWT.BORDER);
		uriCombo.setLayoutData(new GridData(GridData.GRAB_HORIZONTAL | GridData.HORIZONTAL_ALIGN_FILL));
		uriCombo.addModifyListener(new ModifyListener() {
			public void modifyText(ModifyEvent e) {
                validateInput();
			}
		});
        uriCombo.addSelectionListener(new SelectionListener() {
            public void widgetSelected(SelectionEvent e) {
                validateInput();
            }
            public void widgetDefaultSelected(SelectionEvent e) {
                // ignore it
            }
        });
        uriCombo.addMouseListener(new MouseListener() {
            public void mouseUp(MouseEvent e) {
                validateInput();
            }
            public void mouseDown(MouseEvent e) {
                validateInput();
            }
            public void mouseDoubleClick(MouseEvent e) {
                // ignore the event
            }
        });
        uriCombo.addFocusListener(new FocusListener() {
            public void focusLost(FocusEvent e) {
                validateInput();
            }
            public void focusGained(FocusEvent e) {
                validateInput();
            }
        });
		loadBranchesInCombo();
	}

	@Override
	protected Control createDialogArea(Composite parent) {
		Composite composite = (Composite)super.createDialogArea(parent);
		createOptionsArea(composite);
		return composite;
	}

	protected void createOptionsArea(Composite composite) {
		if(optionRemember) {
			remember = new Button(composite, SWT.CHECK);
			remember.setText("Remember");
			remember.setSelection(false); // Start unselected.
		}
		if(optionForce) {
			force = new Button(composite, SWT.CHECK);
			force.setText(UITexts.Dialog_optionForce);
			force.setToolTipText(UITexts.SwitchAction_force_tooltip);
			force.setSelection(false); // Start unselected.
		}
		if(optionUseExistingDir) {
			useExistingDir = new Button(composite, SWT.CHECK);
			useExistingDir.setText("Use existing directory");
			useExistingDir.setToolTipText("Use existing directory");
			useExistingDir.setSelection(false); // Start unselected.
		}
	}

	@Override
	public Control getText() {
		return uriCombo;
	}

	@Override
	public String getTextInput() {
		return uriCombo.getText();
	}

	@Override
	public void setTextInput(String text) {
		uriCombo.setText(text);
	}

	protected Boolean isRemember() {
		return remember.getSelection();
	}

	protected Boolean isForce() {
		return force.getSelection();
	}

	protected boolean isRevisionSelected() {
		return revision.isRevisionSelected(); 
	}
	
	protected String getSelectedRevision() {
		return revision.getRevision(); 
	}

	protected void createRevisionSelector(Group revisionGroup) {
		revision = new RevisionNumber(revisionGroup);
		revision.addShowLogButtonSelectionListener(new SelectionAdapter() {
			@Override
			public void widgetSelected(SelectionEvent e) {
				showLog();
			}
		});
	}
	
	protected void showLog() {
		String url = getTextInput();
		if(url == null || "".equals(url.trim())) {
			return;
		}
        IBzrBranch branch = BranchFactory.findOrCreateFor(url);
		if(branch != null) {
			ShowLogDialog dialog = new ShowLogDialog(getShell(), branch);
			if(dialog.open() == Dialog.OK && dialog.getSelectedRevision() != null) {
				revision.setRevision(dialog.getSelectedRevision().getValue());
			}
		}
	}
	
	@Override
	protected void okPressed() {
		if(optionRemember) {
			info.setRemember(remember.getSelection());
		}
		if(optionForce) {
			info.setForce(force.getSelection());
		}
		if(optionUseExistingDir) {
			info.setUseExistingDir(useExistingDir.getSelection());
		}
		info.setLocation(getValue());
		super.okPressed();
	}

	public static class LocationDialogInfo {
		boolean remember;
		boolean force;
		boolean useExistingDir;
		String location;

		public boolean isRemember() {
			return remember;
		}
		public void setRemember(boolean remember) {
			this.remember = remember;
		}
		public String getLocation() {
			return location;
		}
		public void setLocation(String location) {
			this.location = location;
		}
		public boolean isForce() {
			return force;
		}
		public void setForce(boolean force) {
			this.force = force;
		}
		public boolean isUseExistingDir() {
			return useExistingDir;
		}
		public void setUseExistingDir(boolean useExistingDir) {
			this.useExistingDir = useExistingDir;
		}
	}

    // this main is for test the dialog during development 
    public static void main(String[] args) throws BazaarClientException {
        Display display = new Display();
        Shell shell = new Shell(display);
        shell.setMinimumSize(280, 400);
        LocationDialogInfo info = new LocationDialogInfo();
        List<IBzrBranch> locations = new ArrayList<IBzrBranch>();
        IBzrBranch b = new RemoteBranch(new BranchLocation("lp:bzr-eclipse"));
        locations.add(b);
        LocationInputDialog dialog =
            LocationInputDialog.remotePullLocationDialog(shell, locations, null, info);
        dialog.setBlockOnOpen(true);
        dialog.open();
    }
}
