/*******************************************************************************
 * Copyright (c) 2013 Piotr Piastucki
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package org.vcs.bazaar.eclipse.ui.history;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.IJobChangeEvent;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.core.runtime.jobs.JobChangeAdapter;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.ui.IWorkbenchSite;
import org.eclipse.ui.progress.UIJob;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.BazaarStatusKind;
import org.vcs.bazaar.client.IBazaarLogMessage;
import org.vcs.bazaar.client.IBazaarStatus;
import org.vcs.bazaar.eclipse.core.commands.LogCommand;
import org.vcs.bazaar.eclipse.core.model.IBzrBranch;
import org.vcs.bazaar.eclipse.core.model.IBzrResource;
import org.vcs.bazaar.eclipse.core.resources.IStorageBazaarRevision;
import org.vcs.bazaar.eclipse.internal.core.BazaarException;
import org.vcs.bazaar.eclipse.ui.EclipseBazaarUI;
import org.vcs.bazaar.eclipse.ui.UITexts;
import org.vcs.bazaar.eclipse.ui.actions.AbstractCompareAction;
import org.vcs.bazaar.eclipse.ui.actions.OpenBazaarRevisionAction;
import org.vcs.bazaar.eclipse.ui.operations.DiffOperation;

/**
 * @author Piotr Piastucki
 */
public class AffectedPathsViewer extends TableViewer {

	private IWorkbenchSite site;
	private final IBzrResource resource;
	private final Map<BazaarRevision, List<AffectedPathsItem>> cache;
	
	public AffectedPathsViewer(Composite parent, IWorkbenchSite site, IBzrResource resource) {
		super(parent);
		this.site = site;
		this.resource = resource;
		this.cache = new ConcurrentHashMap<BazaarRevision, List<AffectedPathsItem>>();
		getTable().setLinesVisible(true);
		setLabelProvider(new AffectedPathsLabelProvider());
		setContentProvider(new ArrayContentProvider());
		setComparator(new ViewerComparator() {
			@Override
			public int compare(Viewer viewer, Object element1, Object element2) {
				TableViewer tableViewer = (TableViewer) viewer;
				ITableLabelProvider labelProvider = (ITableLabelProvider) tableViewer.getLabelProvider();
				String l1 = labelProvider.getColumnText(element1, 0);
				String l2 = labelProvider.getColumnText(element2, 0);
				int result = l1.compareTo(l2);
				return result;
			}			
		});
		if (resource.getBranch().isProjectBranch()) {
			final MenuManager mgr = new MenuManager();
			Control c = getControl();
			c.setMenu(mgr.createContextMenu(c));
			final OpenBazaarRevisionAction openAction = new OpenBazaarRevisionAction(UITexts.ActionOpenInEditor_label, site);
			final AbstractCompareAction compareWithPreviousAction = new AbstractCompareAction(UITexts.ActionCompareWithPrevious_label) {
				@Override
				public void internalRun() {
					try {
						IStorageBazaarRevision storage = getStorage(0);
						if (storage.getResource().getType() == IResource.FILE) {
							BazaarRevision previousRev = BazaarRevision.getRevision(BazaarRevision.Prefix.BEFORE, storage.getRevision().toString());
							DiffOperation.openEditor(storage, new IStorageBazaarRevision(storage.getResource(), previousRev));
						}
					} catch (CoreException e) {
						EclipseBazaarUI.log(e);
					}
				}
				@Override
				protected boolean shouldShow(IStructuredSelection selection) {
					if (selection.size() == 1) {
						AffectedPathsItem item = (AffectedPathsItem) selection.getFirstElement();
						return item.getStatus().contains(BazaarStatusKind.MODIFIED);
					}
					return false;
				}
			};
	
			addSelectionChangedListener(new ISelectionChangedListener() {
				public void selectionChanged(SelectionChangedEvent event) {
					IStructuredSelection selection = (IStructuredSelection) getSelection();
					openAction.selectionChanged(selection);
					compareWithPreviousAction.selectionChanged(selection);
				}
			});
			mgr.add(openAction);
			mgr.add(compareWithPreviousAction);
		}
	}
	
	public void setRevision(final BazaarRevision revision) {
		if (revision == null) {
			setInput(null);
			refresh();
		} else {
			if (cache.get(revision) == null) {
				final List<AffectedPathsItem> input = new ArrayList<AffectedPathsItem>();
				Job job = new Job("Fetching affected paths...") {
					@Override
					protected IStatus run(IProgressMonitor monitor) {
						LogCommand lcmd = new LogCommand(resource, revision);
						try {
							lcmd.run(monitor);
							List<IBazaarLogMessage> messages = lcmd.getLogMessages();
							IBzrBranch branch = resource.getBranch();
							if (!messages.isEmpty()) {
								IBazaarLogMessage message = messages.get(0);
								for (IBazaarStatus status : message.getAffectedFiles()) {
									input.add(new AffectedPathsItem(branch, status, revision));
								}
							}
						} catch (BazaarException e) {
						}
						return Status.OK_STATUS;
					}
				};
				job.addJobChangeListener(new JobChangeAdapter() {
					@Override
	                public void done(IJobChangeEvent event) {
						if(event.getResult().isOK()) {
	                        site.getShell().getDisplay().asyncExec(new Runnable() {
								public void run() {
									cache.put(revision, input);
									setInput(input);
									refresh();
								}
							});
	                    }
					}
				});
				job.setUser(true);
				job.setPriority(UIJob.LONG);
				job.schedule();
			} else {
				setInput(cache.get(revision));
				refresh();
			}
		}
	}

}
