/*******************************************************************************
 * Copyright (c) 2012 Idereal Limited
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Alexander Taler
 *******************************************************************************/
package org.vcs.bazaar.eclipse.ui.operations;

import java.io.File;
import java.io.FileFilter;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.operation.IRunnableContext;
import org.eclipse.team.ui.TeamOperation;
import org.vcs.bazaar.client.BazaarRevision;
import org.vcs.bazaar.client.commandline.commands.options.Option;
import org.vcs.bazaar.client.commandline.syntax.ICheckoutOptions;
import org.vcs.bazaar.eclipse.core.commands.BranchCommand;
import org.vcs.bazaar.eclipse.core.commands.CheckoutCommand;
import org.vcs.bazaar.eclipse.core.model.IBzrBranch;
import org.vcs.bazaar.eclipse.core.repository.BazaarTeamProviderType;
import org.vcs.bazaar.eclipse.core.repository.BranchLocationManager;
import org.vcs.bazaar.eclipse.internal.core.BazaarException;

/**
 * Operation for creating project(s) from a Bazaar branch.
 * 
 * Influenced by earlier implementation by Guillermo Gonzalez.
 * 
 * @author Alexander Taler alex <@> idereal <.> co <.> nz
 */
public class CreateBazaarProjectOperation extends TeamOperation {

    public static int TYPE_CHECKOUT = 1;
    public static int TYPE_LIGHTWEIGHT = 2;
    public static int TYPE_BRANCH = 3;

    String projectName;
    IBzrBranch remoteBranch;
    String projectLocation;
    int type;
    String revision;
    
    /**
     * 
     * @param context  Found with getContainer()
     * @param projectName  The project to be created
     * @param remoteBranch  The branch to get code from
     * @param projectLocation Where on disk to check out the project
     * @param type The type of operation, one of TYPE_CHECKOUT, TYPE_LIGHTWEIGHT, TYPE_BRANCH
     * @param revision Optional revision, null means use tip
     */
    public CreateBazaarProjectOperation(IRunnableContext context,
        String projectName,
        IBzrBranch remoteBranch,
        String projectLocation,
        int type,
        String revision)
    {
        super(context);
        this.projectName = projectName;
        this.remoteBranch = remoteBranch;
        this.projectLocation = projectLocation;
        this.type = type;
        this.revision = revision;
    }

    public void run(IProgressMonitor monitor)
        throws InvocationTargetException, InterruptedException
    {
        try {
            if (type == TYPE_BRANCH) {
                BranchCommand op = new BranchCommand(remoteBranch, new File(projectLocation), revision);
                op.run(monitor);
            } else {
                List<Option> options = new ArrayList<Option>(0);
                if (type == TYPE_LIGHTWEIGHT) {
                    options.add(ICheckoutOptions.LIGHTWEIGHT);
                }
                if ((revision != null) && revision.trim().length() > 0) {
                    String s = BazaarRevision.getRevision(BazaarRevision.Prefix.REVNO, revision).toString();
                    options.add(ICheckoutOptions.REVISION.setArgument(s));
                }
                CheckoutCommand op = new CheckoutCommand(remoteBranch, new File(projectLocation), options.toArray(new Option[0]));
                op.run(monitor);
            }

            // Search the resulting checked out directory for existing Eclipse projects.
            List<File> projDirs = findProjects(new File(projectLocation));

            // If no projects were found, then just create a single project at the root.
            if (projDirs.isEmpty()) {
                projDirs.add(new File(projectLocation));
            }

            // Go through the list of project directories and add them all
            for (File projDir : projDirs) {
                createProject(projDir, monitor);
            }
        } catch (BazaarException e) {
            throw new InvocationTargetException(e);
        } catch (CoreException e) {
            throw new InvocationTargetException(e);
        }
    }

    /**
     * Search the given location for Eclipse projects.
     * 
     * This is done after the location has been fetched from Bazaar. The
     * function performs the search recursively, stopping when it finds a
     * project file.
     * 
     * @return List of File objects, the directories containing Bazaar projects.
     */
    private List<File> findProjects(File location) {
        List<File> ps = new LinkedList<File>();
        File dotProj = new File(location, ".project");
        if (dotProj.exists()) {
            ps.add(location);
            return ps;
        }
        // Search directories
        File[] dirs = location.listFiles(new FileFilter() {
            public boolean accept(File pathname) {
                return pathname.isDirectory() && !pathname.getName().equals(".bzr");
            }
        });

        for (File dir : dirs) {
            ps.addAll(findProjects(dir));
        }

        return ps;
    }
    
    /**
     * Create an Eclipse project at the given location.
     *
     * Eclipse throws an exception if a ProjectDescription is used to create
     * a project at a location directly in the workspace root. However, if
     * the project location has the same name as the project itself, then
     * creating the project without a description will not throw an
     * exception.
     * 
     * So, as a special case if the directory is in the workspace root, but
     * matches the project name, create is used without a description.
     * 
     * The other error case is not checked for here, in case this
     * undocumented Eclipse behaviour changes.
     */
    private void createProject(File projDir, IProgressMonitor monitor) throws CoreException {
        IWorkspace ws = ResourcesPlugin.getWorkspace();
        IPath projPath = new Path(projDir.toString());
        IPath parentPath = projPath.removeLastSegments(1);
        IProject projectHandle = null;

        if (parentPath.equals(ws.getRoot().getLocation())
            && projPath.lastSegment().equals(projectName)) {

            projectHandle = ws.getRoot().getProject(projectName);
            projectHandle.create(monitor);

        } else {

            // Derive a new project name based on the directory
            String newProjectName = projectName + "-" + projPath.lastSegment();
            projectHandle = ws.getRoot().getProject(newProjectName);
            IProjectDescription projDesc = ws.newProjectDescription(newProjectName);
            projDesc.setLocation(projPath);
            projectHandle.create(projDesc, monitor);
        }
        projectHandle.open(monitor);
        BazaarTeamProviderType.registerProject(projectHandle);
        BranchLocationManager.getInstance().addRepoLocation(remoteBranch);
    }
}
