/*******************************************************************************
 * Copyright (c) 2012 Idereal Limited
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Alexander Taler
 *******************************************************************************/
package org.vcs.bazaar.eclipse.ui.wizards.createproject;

import java.lang.reflect.InvocationTargetException;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.INewWizard;
import org.eclipse.ui.IWorkbench;
import org.vcs.bazaar.eclipse.BzrWorkspaceRoot;
import org.vcs.bazaar.eclipse.EclipseBazaarCore;
import org.vcs.bazaar.eclipse.core.model.IBzrBranch;
import org.vcs.bazaar.eclipse.core.repository.BranchFactory;
import org.vcs.bazaar.eclipse.internal.core.BazaarException;
import org.vcs.bazaar.eclipse.ui.EclipseBazaarUI;
import org.vcs.bazaar.eclipse.ui.operations.CreateBazaarProjectOperation;
import org.vcs.bazaar.eclipse.ui.team.BazaarUtilities;

/**
 * A Wizard for creating a local Eclipse project from a remote Bazaar branch.
 * 
 * It can be called from the create project and import project menus. It can
 * also be called from the Branch action, in which case the selection is used to
 * determine the remote Bazaar branch.
 * 
 * @author Alexander Taler alex <@> idereal <.> co <.> nz
 */
public class CreateProjectWizard extends Wizard implements INewWizard {

    private CreateProjectPage cpPage;
    private CreateProjectAdvancedPage cpaPage;

    // Initial branch
    protected String initialBranch = null;

    public CreateProjectWizard() {
        super();
    }

    public CreateProjectWizard(IStructuredSelection ss) {
        super();
        init(null, null);

        if ((ss == null) || (ss.size() < 0)) {
            throw new RuntimeException("Selection required");
        }
        if (ss.size() > 1) {
            throw new RuntimeException("Single Selection only supported");
        }
        if (!(ss.getFirstElement() instanceof IProject)) {
            throw new RuntimeException("Selection must be a Project");
        }
        IProject project = (IProject) ss.getFirstElement();
        IBzrBranch branch = BzrWorkspaceRoot.getBzrWorkspaceRootFor(project).getBranch();
        initialBranch = branch.getBranchLocation().toString();
    }

    /**
     * The branch that this wizard was created with.
     */
    String getInitialBranch() {
        return initialBranch;
    }

    /**
     * Display this wizard to the user.
     * 
     * @param shell
     */
    public void show(Shell shell) {
        new WizardDialog(shell, this).open();
    }

	public void init(IWorkbench workbench, IStructuredSelection selection) {
        setWindowTitle("Bazaar Create Project Wizard");
        setNeedsProgressMonitor(true);
        cpPage = new CreateProjectPage("BazaarCreateProjectPage", this);
        cpaPage = new CreateProjectAdvancedPage("BazaarCreateProjectPage", this);

		if (!EclipseBazaarCore.getDefault().isClientAvailable()) {
			BazaarUtilities.configureExecutable();
		}
	}
	
	@Override
	public void addPages() {
        addPage(cpPage);
        addPage(cpaPage);
	}
	

    /**
     * Display a useful error message to the user.
     * 
     * The message is shown on both pages of the wizard.
     * 
     * @param e The exception to show to the user.
     */
    protected void setErrorMessage(Throwable e) {
        String message = "";
        if (e instanceof InvocationTargetException) {
            e = e.getCause();
        }
        if (e instanceof BazaarException) {
            message = ((BazaarException) e).getStatus().getMessage();
        } else if (e instanceof CoreException) {
            if (((CoreException) e).getStatus() != null) {
                message = ((CoreException) e).getStatus().getMessage();
                for (IStatus i : ((CoreException) e).getStatus().getChildren()) {
                    message += i.getMessage();
                }
            } else {
                message = e.getMessage();
            }
        }
        EclipseBazaarUI.log(e);
        cpPage.setErrorMessage(message);
        cpaPage.setErrorMessage(message);
	}

    /**
     * Return a branch object for the branch this wizard will use.
     * 
     * @return A branch object of the appropriate type
     * @throws BazaarException if the user specified branch is invalid
     */
    public IBzrBranch getRemoteBranch() throws BazaarException {
    	return BranchFactory.findOrCreateFor(cpPage.getRepoLocation());
    }

    /**
     * Return the project name currently entered.
     * 
     * @return String containing the project name
     */
    public String getProjectName() {
        return cpPage.getProjectName();
    }

    @Override
    public boolean performFinish() {
        try {
            CreateBazaarProjectOperation op =
                new CreateBazaarProjectOperation(
                    this.getContainer(),
                    cpPage.getProjectName(),
                    getRemoteBranch(),
                    cpPage.getProjectLocation(),
                    cpaPage.getBranchType(),
                    cpaPage.getRevision());
            op.run();
            return true;
        } catch (InvocationTargetException e) {
            setErrorMessage(e);
        } catch (InterruptedException e) {
            setErrorMessage(e);
        } catch (BazaarException e) {
            setErrorMessage(e);
        }
        return false;
    }
}

